import {
  Type,
  PropertyKeyType,
} from '@loafer/core';

import {
  Annotation,
  Class,
  Constructor,
  Field,
  Method,
  Parameter,
  Metadata,
} from '@loafer/core/reflect';

import {
  TypeUtil,
} from '@loafer/core/util/TypeUtil';

import { DecoratorType, DecoratorParametersType } from './type';
import { NotSupportedDecoratorError } from './error';

export class Decorator {

  public static register<T extends Annotation>(AnnotationClass: Type<Annotation>, _annotation: T, ...decoratorArgs: any[]) {
    const annotation: Annotation = Object.assign(new AnnotationClass(), _annotation);

    const name: string = AnnotationClass.name;

    const decoratorType: DecoratorType = Decorator.getDecoratorType(decoratorArgs);

    const [target, propertyKey, descriptorOrParameterIndex] = decoratorArgs;

    const clazz: Class = Class._defineClass(TypeUtil.getType(target));
    let field: Field = null;
    let method: Method = null;
    let parameter: Parameter = null;
    let cons: Constructor = null;

    switch (decoratorType) {
    case DecoratorType.CLASS:
      try {
        cons = clazz._defineConstructor(Metadata.getOwnParamTypes(target));
        clazz._addAnnotation(annotation);
      } catch (e) {
        if (e instanceof NotSupportedDecoratorError) {
          throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Class.`);
        }
        throw e;
      }
      break;
    case DecoratorType.PROPERTY:
      try {
        field = clazz._defineField(propertyKey, Metadata.getOwnType(target, propertyKey));
        field._addAnnotation(annotation);
      } catch (e) {
        if (e instanceof NotSupportedDecoratorError) {
          throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Property.`);
        }
        throw e;
      }
      break;
    case DecoratorType.METHOD:
      try {
        method = clazz._defineMethod(propertyKey,
                                     Metadata.getOwnParamTypes(target, propertyKey),
                                     Metadata.getOwnReturnType(target, propertyKey));
        method._addAnnotation(annotation);
      } catch (e) {
        if (e instanceof NotSupportedDecoratorError) {
          throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Method.`);
        }
        throw e;
      }
      break;
    case DecoratorType.PARAMETER:
      try {
        if (undefined === propertyKey) {
          cons = clazz.getConstructor();
          parameter = cons.getParameter(descriptorOrParameterIndex);
        } else {
          method = clazz._defineMethod(propertyKey,
                                       Metadata.getOwnParamTypes(target, propertyKey),
                                       Metadata.getOwnReturnType(target, propertyKey));
          parameter = method.getParameter(descriptorOrParameterIndex);
        }
        parameter._addAnnotation(annotation);
      } catch (e) {
        if (e instanceof NotSupportedDecoratorError) {
          throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Parameter.`);
        }
        throw e;
      }
      break;
    default:
      throw new NotSupportedDecoratorError(`Cannot determine decorator[@${name}] type.`);
    }
  }

  /**
   *
   * @param {any[]} args
   * @returns {DecoratorType}
   */
  public static getDecoratorType(args: any[]): DecoratorType {
    const [, propertyKey, descriptor] = args;

    if (typeof descriptor === 'number') {
        return DecoratorType.PARAMETER;
    }

    if (propertyKey && descriptor === undefined || descriptor && (descriptor.get || descriptor.set)) {
        return DecoratorType.PROPERTY;
    }
    return (descriptor && descriptor.value) ? DecoratorType.METHOD : DecoratorType.CLASS;
  }

  /**
   *
   * @param target
   * @param {string} propertyKey
   * @returns {DecoratorParametersType}
   */
  public static decoratorArgs(target: any, propertyKey: PropertyKeyType): DecoratorParametersType {
    return [
        target,
        propertyKey,
        TypeUtil.descriptorOf(target, propertyKey),
    ];
  }
}
